# Based on https://github.com/NVIDIA/TensorRT/blob/main/demo/Diffusion/models.py
#
#
# SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Union, List, Optional, Any, Tuple

from nebullvm.core.models import Device
from nebullvm.optional_modules.diffusers import (
    DiffusionPipeline,
    UNet2DConditionModel,
    UNet2DOutput,
    AutoencoderKL,
    onnx_graphsurgeon as gs,
)
from nebullvm.optional_modules.diffusers import StableDiffusionPipeline
from nebullvm.optional_modules.huggingface import CLIPTextModel, CLIPTokenizer
from nebullvm.optional_modules.onnx import onnx
from nebullvm.optional_modules.tensor_rt import fold_constants
from nebullvm.optional_modules.torch import torch


@torch.no_grad()
def get_unet_inputs(
    self,
    prompt: Union[str, List[str]] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
    num_inference_steps: int = 1,
    guidance_scale: float = 7.5,
    negative_prompt: Optional[Union[str, List[str]]] = None,
    num_images_per_prompt: Optional[int] = 1,
    generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
    latents: Optional[torch.FloatTensor] = None,
    prompt_embeds: Optional[torch.FloatTensor] = None,
    negative_prompt_embeds: Optional[torch.FloatTensor] = None,
    callback_steps: int = 1,
    cross_attention_kwargs: Optional[Dict[str, Any]] = None,
):
    # 0. Default height and width to unet
    height = height or self.unet.config.sample_size * self.vae_scale_factor
    width = width or self.unet.config.sample_size * self.vae_scale_factor

    # 1. Check inputs. Raise error if not correct
    self.check_inputs(
        prompt,
        height,
        width,
        callback_steps,
        negative_prompt,
        prompt_embeds,
        negative_prompt_embeds,
    )

    # 2. Define call parameters
    if prompt is not None and isinstance(prompt, str):
        batch_size = 1
    elif prompt is not None and isinstance(prompt, list):
        batch_size = len(prompt)
    else:
        batch_size = prompt_embeds.shape[0]

    device = self._execution_device
    do_classifier_free_guidance = guidance_scale > 1.0

    # 3. Encode input prompt
    prompt_embeds = self._encode_prompt(
        prompt,
        device,
        num_images_per_prompt,
        do_classifier_free_guidance,
        negative_prompt,
        prompt_embeds=prompt_embeds,
        negative_prompt_embeds=negative_prompt_embeds,
    )

    # 4. Prepare timesteps
    self.scheduler.set_timesteps(num_inference_steps, device=device)
    timesteps = self.scheduler.timesteps

    # 5. Prepare latent variables
    num_channels_latents = self.unet.in_channels
    latents = self.prepare_latents(
        batch_size * num_images_per_prompt,
        num_channels_latents,
        height,
        width,
        prompt_embeds.dtype,
        device,
        generator,
        latents,
    )

    for i, t in enumerate(timesteps):
        # expand the latents if we are doing classifier free guidance
        latent_model_input = (
            torch.cat([latents] * 2)
            if do_classifier_free_guidance
            else latents
        )
        latent_model_input = self.scheduler.scale_model_input(
            latent_model_input, t
        )

        return latent_model_input, t, prompt_embeds, cross_attention_kwargs


class DiffusionUNetWrapper(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, *x, **kwargs):
        return tuple(
            self.model(x[0], x[1], encoder_hidden_states=x[2]).values()
        )


class OptimizedDiffusionWrapper(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, *x, **kwargs):
        return UNet2DOutput(
            self.model(
                x[0],
                x[1].reshape((1,)) if x[1].shape == torch.Size([]) else x[1],
                kwargs["encoder_hidden_states"],
            )[0]
        )


def is_diffusion_model_pipe(model):
    return isinstance(model, DiffusionPipeline)


def get_default_dynamic_info(input_shape: List[Tuple[int, ...]]):
    return {
        "inputs": [
            {
                0: {
                    "name": "2B",
                    "min_val": input_shape[0][0],
                    "opt_val": input_shape[0][0],
                    "max_val": input_shape[0][0],
                },
                2: {
                    "name": "H",
                    "min_val": input_shape[0][2],
                    "opt_val": input_shape[0][2],
                    "max_val": input_shape[0][2],
                },
                3: {
                    "name": "W",
                    "min_val": input_shape[0][3],
                    "opt_val": input_shape[0][3],
                    "max_val": input_shape[0][3],
                },
            },
            {},
            {
                0: {
                    "name": "2B",
                    "min_val": input_shape[2][0],
                    "opt_val": input_shape[2][0],
                    "max_val": input_shape[2][0],
                }
            },
        ],
        "outputs": [{0: "2B", 2: "H", 3: "W"}],
    }


def preprocess_diffusers(pipe: DiffusionPipeline) -> torch.nn.Module:
    # Function that wraps the Diffusion UNet model to
    # be compatible with the optimizations performed by nebullvm
    model = DiffusionUNetWrapper(pipe.unet)
    return model


def postprocess_diffusers(
    optimized_model: Any,
    pipe: StableDiffusionPipeline,
    device: Device,
) -> StableDiffusionPipeline:
    # Function that puts the optimized Diffusion UNet model back
    # into the Diffusion Pipeline
    final_model = OptimizedDiffusionWrapper(optimized_model)
    final_model.sample_size = pipe.unet.sample_size
    final_model.in_channels = pipe.unet.in_channels
    final_model.device = torch.device(device.to_torch_format())
    final_model.config = pipe.unet.config
    final_model.in_channels = pipe.unet.in_channels
    pipe.unet = final_model
    return pipe


class Optimizer:
    def __init__(self, onnx_graph, verbose=False):
        self.graph = gs.import_onnx(onnx_graph)
        self.verbose = verbose

    def info(self, prefix):
        if self.verbose:
            print(
                f"{prefix} .. {len(self.graph.nodes)} nodes, {len(self.graph.tensors().keys())} tensors, {len(self.graph.inputs)} inputs, {len(self.graph.outputs)} outputs"
            )

    def cleanup(self, return_onnx=False):
        self.graph.cleanup().toposort()
        if return_onnx:
            return gs.export_onnx(self.graph)

    def select_outputs(self, keep, names=None):
        self.graph.outputs = [self.graph.outputs[o] for o in keep]
        if names:
            for i, name in enumerate(names):
                self.graph.outputs[i].name = name

    def fold_constants(self, return_onnx=False):
        onnx_graph = fold_constants(
            gs.export_onnx(self.graph),
            allow_onnxruntime_shape_inference=True,
        )
        self.graph = gs.import_onnx(onnx_graph)
        if return_onnx:
            return onnx_graph

    def infer_shapes(self, return_onnx=False):
        onnx_graph = gs.export_onnx(self.graph)
        if onnx_graph.ByteSize() > 2147483648:
            raise TypeError("ERROR: model size exceeds supported 2GB limit")
        else:
            onnx_graph = onnx.shape_inference.infer_shapes(onnx_graph)

        self.graph = gs.import_onnx(onnx_graph)
        if return_onnx:
            return onnx_graph


def get_path(version, inpaint=False):
    if version == "1.4":
        if inpaint:
            return "runwayml/stable-diffusion-inpainting"
        else:
            return "CompVis/stable-diffusion-v1-4"
    elif version == "1.5":
        if inpaint:
            return "runwayml/stable-diffusion-inpainting"
        else:
            return "runwayml/stable-diffusion-v1-5"
    elif version == "2.0-base":
        if inpaint:
            return "stabilityai/stable-diffusion-2-inpainting"
        else:
            return "stabilityai/stable-diffusion-2-base"
    elif version == "2.0":
        if inpaint:
            return "stabilityai/stable-diffusion-2-inpainting"
        else:
            return "stabilityai/stable-diffusion-2"
    elif version == "2.1":
        return "stabilityai/stable-diffusion-2-1"
    elif version == "2.1-base":
        return "stabilityai/stable-diffusion-2-1-base"
    else:
        raise ValueError(f"Incorrect version {version}")


def get_embedding_dim(version):
    if version in ("1.4", "1.5"):
        return 768
    elif version in ("2.0", "2.0-base", "2.1", "2.1-base"):
        return 1024
    else:
        raise ValueError(f"Incorrect version {version}")


class BaseModel:
    def __init__(
        self,
        hf_token,
        fp16=False,
        device="cuda",
        verbose=False,
        path="",
        max_batch_size=16,
        embedding_dim=768,
        text_maxlen=77,
    ):
        self.name = "SD Model"
        self.hf_token = hf_token
        self.fp16 = fp16
        self.device = device
        self.verbose = verbose
        self.path = path

        self.min_batch = 1
        self.max_batch = max_batch_size
        self.min_image_shape = 256  # min image resolution: 256x256
        self.max_image_shape = 1024  # max image resolution: 1024x1024
        self.min_latent_shape = self.min_image_shape // 8
        self.max_latent_shape = self.max_image_shape // 8

        self.embedding_dim = embedding_dim
        self.text_maxlen = text_maxlen

    def get_model(self):
        pass

    def get_input_names(self):
        pass

    def get_output_names(self):
        pass

    def get_dynamic_axes(self):
        return None

    def get_sample_input(self, batch_size, image_height, image_width):
        pass

    def get_input_profile(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        return None

    def get_shape_dict(self, batch_size, image_height, image_width):
        return None

    def optimize(self, onnx_graph):
        opt = Optimizer(onnx_graph, verbose=self.verbose)
        opt.info(self.name + ": original")
        opt.cleanup()
        opt.info(self.name + ": cleanup")
        opt.fold_constants()
        opt.info(self.name + ": fold constants")
        opt.infer_shapes()
        opt.info(self.name + ": shape inference")
        onnx_opt_graph = opt.cleanup(return_onnx=True)
        opt.info(self.name + ": finished")
        return onnx_opt_graph

    def check_dims(self, batch_size, image_height, image_width):
        assert batch_size >= self.min_batch and batch_size <= self.max_batch
        assert image_height % 8 == 0 or image_width % 8 == 0
        latent_height = image_height // 8
        latent_width = image_width // 8
        assert (
            latent_height >= self.min_latent_shape
            and latent_height <= self.max_latent_shape
        )
        assert (
            latent_width >= self.min_latent_shape
            and latent_width <= self.max_latent_shape
        )
        return (latent_height, latent_width)

    def get_minmax_dims(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        min_batch = batch_size if static_batch else self.min_batch
        max_batch = batch_size if static_batch else self.max_batch
        latent_height = image_height // 8
        latent_width = image_width // 8
        min_image_height = (
            image_height if static_shape else self.min_image_shape
        )
        max_image_height = (
            image_height if static_shape else self.max_image_shape
        )
        min_image_width = image_width if static_shape else self.min_image_shape
        max_image_width = image_width if static_shape else self.max_image_shape
        min_latent_height = (
            latent_height if static_shape else self.min_latent_shape
        )
        max_latent_height = (
            latent_height if static_shape else self.max_latent_shape
        )
        min_latent_width = (
            latent_width if static_shape else self.min_latent_shape
        )
        max_latent_width = (
            latent_width if static_shape else self.max_latent_shape
        )
        return (
            min_batch,
            max_batch,
            min_image_height,
            max_image_height,
            min_image_width,
            max_image_width,
            min_latent_height,
            max_latent_height,
            min_latent_width,
            max_latent_width,
        )


class CLIP(BaseModel):
    def __init__(
        self, hf_token, device, verbose, path, max_batch_size, embedding_dim
    ):
        super(CLIP, self).__init__(
            hf_token,
            device=device,
            verbose=verbose,
            path=path,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
        )
        self.name = "CLIP"

    def get_model(self):
        return CLIPTextModel.from_pretrained(
            self.path, subfolder="text_encoder", use_auth_token=self.hf_token
        ).to(self.device)

    def get_input_names(self):
        return ["input_ids"]

    def get_output_names(self):
        return ["text_embeddings", "pooler_output"]

    def get_dynamic_axes(self):
        return {"input_ids": {0: "B"}, "text_embeddings": {0: "B"}}

    def get_input_profile(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        self.check_dims(batch_size, image_height, image_width)
        min_batch, max_batch, _, _, _, _, _, _, _, _ = self.get_minmax_dims(
            batch_size, image_height, image_width, static_batch, static_shape
        )
        return {
            "input_ids": [
                (min_batch, self.text_maxlen),
                (batch_size, self.text_maxlen),
                (max_batch, self.text_maxlen),
            ]
        }

    def get_shape_dict(self, batch_size, image_height, image_width):
        self.check_dims(batch_size, image_height, image_width)
        return {
            "input_ids": (batch_size, self.text_maxlen),
            "text_embeddings": (
                batch_size,
                self.text_maxlen,
                self.embedding_dim,
            ),
        }

    def get_sample_input(self, batch_size, image_height, image_width):
        self.check_dims(batch_size, image_height, image_width)
        return torch.zeros(
            batch_size, self.text_maxlen, dtype=torch.int32, device=self.device
        )

    def optimize(self, onnx_graph):
        opt = Optimizer(onnx_graph, verbose=self.verbose)
        opt.info(self.name + ": original")
        opt.select_outputs([0])  # delete graph output#1
        opt.cleanup()
        opt.info(self.name + ": remove output[1]")
        opt.fold_constants()
        opt.info(self.name + ": fold constants")
        opt.infer_shapes()
        opt.info(self.name + ": shape inference")
        opt.select_outputs(
            [0], names=["text_embeddings"]
        )  # rename network output
        opt.info(self.name + ": remove output[0]")
        opt_onnx_graph = opt.cleanup(return_onnx=True)
        opt.info(self.name + ": finished")
        return opt_onnx_graph


def make_CLIP(
    version, hf_token, device, verbose, max_batch_size, inpaint=False
):
    return CLIP(
        hf_token=hf_token,
        device=device,
        verbose=verbose,
        path=get_path(version, inpaint=inpaint),
        max_batch_size=max_batch_size,
        embedding_dim=get_embedding_dim(version),
    )


class UNet(BaseModel):
    def __init__(
        self,
        hf_token,
        fp16=False,
        device="cuda",
        verbose=False,
        path="",
        max_batch_size=16,
        embedding_dim=768,
        text_maxlen=77,
        unet_dim=4,
    ):
        super(UNet, self).__init__(
            hf_token,
            fp16=fp16,
            device=device,
            verbose=verbose,
            path=path,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
            text_maxlen=text_maxlen,
        )
        self.unet_dim = unet_dim
        self.name = "UNet"

    def get_model(self):
        model_opts = (
            {"revision": "fp16", "torch_dtype": torch.float16}
            if self.fp16
            else {}
        )
        return UNet2DConditionModel.from_pretrained(
            self.path,
            subfolder="unet",
            use_auth_token=self.hf_token,
            **model_opts,
        ).to(self.device)

    def get_input_names(self):
        return ["sample", "timestep", "encoder_hidden_states"]

    def get_output_names(self):
        return ["latent"]

    def get_dynamic_axes(self):
        return {
            "sample": {0: "2B", 2: "H", 3: "W"},
            "encoder_hidden_states": {0: "2B"},
            "latent": {0: "2B", 2: "H", 3: "W"},
        }

    def get_input_profile(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        (
            min_batch,
            max_batch,
            _,
            _,
            _,
            _,
            min_latent_height,
            max_latent_height,
            min_latent_width,
            max_latent_width,
        ) = self.get_minmax_dims(
            batch_size, image_height, image_width, static_batch, static_shape
        )
        return {
            "sample": [
                (
                    2 * min_batch,
                    self.unet_dim,
                    min_latent_height,
                    min_latent_width,
                ),
                (2 * batch_size, self.unet_dim, latent_height, latent_width),
                (
                    2 * max_batch,
                    self.unet_dim,
                    max_latent_height,
                    max_latent_width,
                ),
            ],
            "encoder_hidden_states": [
                (2 * min_batch, self.text_maxlen, self.embedding_dim),
                (2 * batch_size, self.text_maxlen, self.embedding_dim),
                (2 * max_batch, self.text_maxlen, self.embedding_dim),
            ],
        }

    def get_shape_dict(self, batch_size, image_height, image_width):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        return {
            "sample": (
                2 * batch_size,
                self.unet_dim,
                latent_height,
                latent_width,
            ),
            "encoder_hidden_states": (
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
            ),
            "latent": (2 * batch_size, 4, latent_height, latent_width),
        }

    def get_sample_input(self, batch_size, image_height, image_width):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        dtype = torch.float16 if self.fp16 else torch.float32
        return (
            torch.randn(
                2 * batch_size,
                self.unet_dim,
                latent_height,
                latent_width,
                dtype=torch.float32,
                device=self.device,
            ),
            torch.tensor([1.0], dtype=torch.float32, device=self.device),
            torch.randn(
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
                dtype=dtype,
                device=self.device,
            ),
        )


def make_UNet(
    version, hf_token, device, verbose, max_batch_size, inpaint=False
):
    return UNet(
        hf_token=hf_token,
        fp16=True,
        device=device,
        verbose=verbose,
        path=get_path(version, inpaint=inpaint),
        max_batch_size=max_batch_size,
        embedding_dim=get_embedding_dim(version),
        unet_dim=(9 if inpaint else 4),
    )


class VAE(BaseModel):
    def __init__(
        self, hf_token, device, verbose, path, max_batch_size, embedding_dim
    ):
        super(VAE, self).__init__(
            hf_token,
            device=device,
            verbose=verbose,
            path=path,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
        )
        self.name = "VAE decoder"

    def get_model(self):
        vae = AutoencoderKL.from_pretrained(
            self.path, subfolder="vae", use_auth_token=self.hf_token
        ).to(self.device)
        vae.forward = vae.decode
        return vae

    def get_input_names(self):
        return ["latent"]

    def get_output_names(self):
        return ["images"]

    def get_dynamic_axes(self):
        return {
            "latent": {0: "B", 2: "H", 3: "W"},
            "images": {0: "B", 2: "8H", 3: "8W"},
        }

    def get_input_profile(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        (
            min_batch,
            max_batch,
            _,
            _,
            _,
            _,
            min_latent_height,
            max_latent_height,
            min_latent_width,
            max_latent_width,
        ) = self.get_minmax_dims(
            batch_size, image_height, image_width, static_batch, static_shape
        )
        return {
            "latent": [
                (min_batch, 4, min_latent_height, min_latent_width),
                (batch_size, 4, latent_height, latent_width),
                (max_batch, 4, max_latent_height, max_latent_width),
            ]
        }

    def get_shape_dict(self, batch_size, image_height, image_width):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        return {
            "latent": (batch_size, 4, latent_height, latent_width),
            "images": (batch_size, 3, image_height, image_width),
        }

    def get_sample_input(self, batch_size, image_height, image_width):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        return torch.randn(
            batch_size,
            4,
            latent_height,
            latent_width,
            dtype=torch.float32,
            device=self.device,
        )


def make_VAE(
    version, hf_token, device, verbose, max_batch_size, inpaint=False
):
    return VAE(
        hf_token=hf_token,
        device=device,
        verbose=verbose,
        path=get_path(version, inpaint=inpaint),
        max_batch_size=max_batch_size,
        embedding_dim=get_embedding_dim(version),
    )


class TorchVAEEncoder(torch.nn.Module):
    def __init__(self, token, device, path):
        super().__init__()
        self.path = path
        self.vae_encoder = AutoencoderKL.from_pretrained(
            self.path, subfolder="vae", use_auth_token=token
        ).to(device)

    def forward(self, x):
        return self.vae_encoder.encode(x).latent_dist.sample()


class VAEEncoder(BaseModel):
    def __init__(
        self, hf_token, device, verbose, path, max_batch_size, embedding_dim
    ):
        super(VAEEncoder, self).__init__(
            hf_token,
            device=device,
            verbose=verbose,
            path=path,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
        )
        self.name = "VAE encoder"

    def get_model(self):
        vae_encoder = TorchVAEEncoder(self.hf_token, self.device, self.path)
        return vae_encoder

    def get_input_names(self):
        return ["images"]

    def get_output_names(self):
        return ["latent"]

    def get_dynamic_axes(self):
        return {
            "images": {0: "B", 2: "8H", 3: "8W"},
            "latent": {0: "B", 2: "H", 3: "W"},
        }

    def get_input_profile(
        self, batch_size, image_height, image_width, static_batch, static_shape
    ):
        assert batch_size >= self.min_batch and batch_size <= self.max_batch
        min_batch = batch_size if static_batch else self.min_batch
        max_batch = batch_size if static_batch else self.max_batch
        self.check_dims(batch_size, image_height, image_width)
        (
            min_batch,
            max_batch,
            min_image_height,
            max_image_height,
            min_image_width,
            max_image_width,
            _,
            _,
            _,
            _,
        ) = self.get_minmax_dims(
            batch_size, image_height, image_width, static_batch, static_shape
        )

        return {
            "images": [
                (min_batch, 3, min_image_height, min_image_width),
                (batch_size, 3, image_height, image_width),
                (max_batch, 3, max_image_height, max_image_width),
            ],
        }

    def get_shape_dict(self, batch_size, image_height, image_width):
        latent_height, latent_width = self.check_dims(
            batch_size, image_height, image_width
        )
        return {
            "images": (batch_size, 3, image_height, image_width),
            "latent": (batch_size, 4, latent_height, latent_width),
        }

    def get_sample_input(self, batch_size, image_height, image_width):
        self.check_dims(batch_size, image_height, image_width)
        return torch.randn(
            batch_size,
            3,
            image_height,
            image_width,
            dtype=torch.float32,
            device=self.device,
        )


def make_VAEEncoder(
    version, hf_token, device, verbose, max_batch_size, inpaint=False
):
    return VAEEncoder(
        hf_token=hf_token,
        device=device,
        verbose=verbose,
        path=get_path(version, inpaint=inpaint),
        max_batch_size=max_batch_size,
        embedding_dim=get_embedding_dim(version),
    )


def make_tokenizer(version, hf_token):
    return CLIPTokenizer.from_pretrained(
        get_path(version), subfolder="tokenizer", use_auth_token=hf_token
    )


def is_diffusion_model(model) -> bool:
    try:
        from diffusers import UNet2DConditionModel
    except ImportError:
        return False

    if is_diffusion_model_pipe(model):
        return True
    if isinstance(model, (UNet2DConditionModel, DiffusionUNetWrapper)):
        return True
    if hasattr(model, "model"):
        return isinstance(model.model, UNet2DConditionModel)
    return False



from nebullvm.optional_modules.dummy import DummyClass

try:
    import absl.logging

    absl.logging.set_verbosity(absl.logging.ERROR)
except Exception:
    pass


class Keras:
    Model = DummyClass


class data:
    Dataset = DummyClass


class dtypes:
    DType = DummyClass


class Tensorflow:
    Module = DummyClass
    Tensor = DummyClass
    keras = Keras()
    data = data
    dtypes = dtypes
    float16 = float32 = int32 = int64 = DummyClass

    @staticmethod
    def function(**kwargs):
        return lambda x: x


try:
    import tensorflow  # noqa F401

    physical_devices = tensorflow.config.experimental.list_physical_devices(
        "GPU"
    )
    if len(physical_devices) > 0:
        for physical_device in physical_devices:
            tensorflow.config.experimental.set_memory_growth(
                physical_device, True
            )

    tensorflow.get_logger().setLevel("ERROR")
    tensorflow.autograph.set_verbosity(0)
except (ImportError, AttributeError):
    tensorflow = Tensorflow


try:
    import tf2onnx  # noqa F401

    tf2onnx.logging.set_level("ERROR")
    tf2onnx.logging.set_tf_verbosity("ERROR")
except ImportError:
    tf2onnx = object



import os
import shutil
from abc import ABC
from pathlib import Path
from typing import Union, List, Generator, Tuple, Dict, Type

import numpy as np
from loguru import logger

from nebullvm.config import ONNX_FILENAMES
from nebullvm.core.models import Device, ModelParams, DeepLearningFramework
from nebullvm.operations.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
)
from nebullvm.optional_modules.deepsparse import cpu, compile_model
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.transformations import MultiStageTransformation


class DeepSparseInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized on CPU using DeepSparse. DeepSparse is an engine
    accelerating sparse computations on CPUs.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    name = "DeepSparse"

    def __init__(
        self,
        onnx_path: Union[str, Path],
        input_names: List[str],
        output_names: List[str],
        device: Device,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.onnx_path = self._store_file(onnx_path)

        # Compile model
        cores_per_socket, _, _ = cpu.cpu_details()
        # Define the number of cores to use, by default it will make use of
        # all physical cores on the system
        num_cores = cores_per_socket
        batch_size = kwargs["network_parameters"].batch_size
        self.engine = compile_model(onnx_path, batch_size, num_cores)

        self.input_names = input_names
        self.output_names = output_names
        self.device = device

    def get_size(self):
        return os.path.getsize(self.onnx_path)

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        metadata = LearnerMetadata.from_model(
            self,
            input_names=self.input_names,
            output_names=self.output_names,
            **kwargs,
        )
        metadata.save(path)

        shutil.copy(
            self.onnx_path,
            Path(path) / ONNX_FILENAMES["model_name"],
        )

    def free_gpu_memory(self):
        raise NotImplementedError("DeepSparse does not support GPU inference.")

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for consistency
                with other Learners.

        Returns:
            DeepSparseInferenceLearner: The optimized model.
        """
        if len(kwargs) > 0:
            logger.warning(
                f"No extra keywords expected for the load method. "
                f"Got {kwargs}."
            )
        onnx_path = os.path.join(str(path), ONNX_FILENAMES["model_name"])
        metadata = LearnerMetadata.read(path)
        input_tfms = metadata.input_tfms
        if input_tfms is not None:
            input_tfms = MultiStageTransformation.from_dict(
                metadata.input_tfms
            )
        device = Device.from_str(metadata.device)
        return cls(
            input_tfms=input_tfms,
            network_parameters=ModelParams(**metadata.network_parameters),
            onnx_path=onnx_path,
            input_names=metadata["input_names"],
            output_names=metadata["output_names"],
            device=device,
        )

    def _predict_arrays(self, input_arrays: Generator[np.ndarray, None, None]):
        inputs = [array for array in input_arrays]
        outputs = self.engine(inputs)
        return outputs


class PytorchDeepSparseInferenceLearner(
    DeepSparseInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized on CPU using DeepSparse. DeepSparse is an engine
    accelerating sparse computations on CPUs.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        outputs = self._predict_arrays(input_arrays)
        return tuple(torch.from_numpy(output) for output in outputs)


DEEPSPARSE_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[DeepSparseInferenceLearner]
] = {DeepLearningFramework.PYTORCH: PytorchDeepSparseInferenceLearner}



from typing import Union, List, Tuple, Any, Optional, Dict

import numpy as np
from loguru import logger

from nebullvm.core.models import Device, DataType, InputInfo
from nebullvm.optional_modules.tensorflow import tensorflow as tf


def get_output_info_tf(
    tf_model: Union[tf.Module, tf.keras.Model],
    input_tensors: List[tf.Tensor],
    device: Device,
) -> List[Tuple[Tuple[int, ...], DataType]]:
    with tf.device(device.to_tf_format()):
        outputs = tf_model(input_tensors)
    if isinstance(outputs, tf.Tensor) and outputs is not None:
        return [
            (
                tuple(outputs.shape),
                DataType.from_framework_format(outputs.dtype),
            )
        ]
    return [
        (tuple(x.shape), DataType.from_framework_format(x.dtype))
        for x in outputs
    ]


def create_model_inputs_tf(input_infos: List[InputInfo]) -> List[tf.Tensor]:
    return [
        tf.random_normal_initializer()(
            shape=(
                input_info.size[0],
                *input_info.size[2:],
                input_info.size[1],
            )
        )
        if input_info.dtype is DataType.FLOAT32
        else tf.random.uniform(
            shape=(
                input_info.size[0],
                *input_info.size[2:],
                input_info.size[1],
            ),
            minval=input_info.min_value or 0,
            maxval=input_info.max_value or 100,
            dtype=tf.int32,
        )
        for input_info in input_infos
    ]


def run_tf_model(
    model: tf.Module,
    input_tensors: Tuple[tf.Tensor],
    device: Device,
) -> Tuple[tf.Tensor]:
    with tf.device(device.to_tf_format()):
        pred = model(input_tensors)
    if isinstance(pred, tf.Tensor):
        pred = (pred,)
    return pred


def _extract_dynamic_axis(
    tf_model: tf.Module,
    dataset: List[Tuple[Tuple[tf.Tensor, ...], Any]],
    input_sizes: List[Tuple[int, ...]],
    device: Device,
    max_data: int = 100,
) -> Optional[Dict]:
    from nebullvm.tools.utils import inspect_dynamic_size

    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, input_data in enumerate(dataset):
        input_tensors = input_data[0]
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, dynamic_axis["inputs"]
        )
        outputs = tuple(run_tf_model(tf_model, input_tensors, device))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(outputs, output_sizes, dynamic_axis["outputs"])
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def extract_info_from_tf_data(
    tf_model: tf.Module,
    dataset: List[Tuple[Tuple[tf.Tensor, ...], Any]],
    dynamic_axis: Dict,
    device: Device,
    **kwargs,
):
    from nebullvm.tools.utils import ifnone

    input_row = dataset[0][0]
    batch_size = int(input_row[0].shape[0])
    if not all([input_row[0].shape[0] == x.shape[0] for x in input_row]):
        logger.warning("Detected not consistent batch size in the inputs.")

    input_sizes = [tuple(x.shape) for x in input_row]
    input_types = [
        "int32"
        if x.dtype in [tf.int32, np.int32]
        else "int64"
        if x.dtype in [tf.int64, np.int64]
        else "float16"
        if x.dtype in [tf.float16, np.float16]
        else "float32"
        for x in input_row
    ]

    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(tf_model, dataset, input_sizes, device),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def tensorflow_is_gpu_available():
    return len(tf.config.list_physical_devices("GPU")) > 0


def tensorflow_get_gpu_name():
    gpu_devices = tf.config.list_physical_devices("GPU")
    if gpu_devices:
        details = tf.config.experimental.get_device_details(gpu_devices[0])
        details.get("device_name", "Unknown GPU")
        return details["device_name"]
    else:
        return "Unknown GPU"



import subprocess
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from typing import Optional, Any, Union, Tuple, List, Dict

import numpy as np

from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch


class DeepLearningFramework(Enum):
    PYTORCH = "torch"
    TENSORFLOW = "tensorflow"
    NUMPY = "numpy"


class QuantizationType(Enum):
    DYNAMIC = "DYNAMIC"
    STATIC = "STATIC"
    HALF = "HALF"


class Status(Enum):
    OK = "OK"
    ERROR = "ERROR"


class DeviceType(Enum):
    CPU = "cpu"
    GPU = "gpu"
    TPU = "tpu"
    NEURON = "neuron"


class DataType(str, Enum):
    FLOAT16 = "float16"
    FLOAT32 = "float32"
    INT32 = "int32"
    INT64 = "int64"

    @classmethod
    def from_framework_format(
        cls, dtype: Union[torch.dtype, tf.dtypes.DType, np.dtype]
    ):
        if isinstance(dtype, torch.dtype):
            framework = "torch"
        elif isinstance(dtype, tf.dtypes.DType):
            framework = "tensorflow"
        else:
            framework = "numpy"
            dtype = dtype.type
        return FRAMEWORK_TO_DATA_TYPE_CONVERSION_DICT[framework][dtype]

    def to_torch_format(self):
        for key, value in FRAMEWORK_TO_DATA_TYPE_CONVERSION_DICT[
            "torch"
        ].items():
            if value == self:
                return key

    def to_tf_format(self):
        for key, value in FRAMEWORK_TO_DATA_TYPE_CONVERSION_DICT[
            "tensorflow"
        ].items():
            if value == self:
                return key

    def to_numpy_format(self):
        for key, value in FRAMEWORK_TO_DATA_TYPE_CONVERSION_DICT[
            "numpy"
        ].items():
            if value == self:
                return key


class ModelCompiler(Enum):
    TENSOR_RT = "tensor_rt"
    TENSOR_RT_ONNX = "onnx_tensor_rt"
    TENSOR_RT_TORCH = "torch_tensor_rt"
    OPENVINO = "openvino"
    APACHE_TVM = "tvm"
    APACHE_TVM_TORCH = "torch_tvm"
    APACHE_TVM_ONNX = "onnx_tvm"
    ONNX_RUNTIME = "onnxruntime"
    DEEPSPARSE = "deepsparse"
    TORCHSCRIPT = "torchscript"
    XLA = "xla"
    TFLITE = "tflite"
    BLADEDISC = "bladedisc"
    INTEL_NEURAL_COMPRESSOR = "intel_neural_compressor"
    TORCH_NEURON = "torch_neuron"
    TORCH_XLA = "torch_xla"
    TORCH_DYNAMO = "torch_dynamo"
    FASTER_TRANSFORMER = "faster_transformer"


class ModelCompressor(Enum):
    SPARSE_ML = "sparseml"
    INTEL_PRUNING = "intel_pruning"


class OptimizationTime(Enum):
    CONSTRAINED = "constrained"
    UNCONSTRAINED = "unconstrained"


@dataclass
class HardwareSetup:
    cpu: str
    operating_system: str
    memory_gb: int
    accelerator: Optional[str] = None


@dataclass
class OptimizedModel:
    inference_learner: Any
    latency_seconds: float
    metric_drop: float
    technique: str
    compiler: str
    throughput: float
    size_mb: float


@dataclass
class OriginalModel:
    model: Any
    latency_seconds: float
    throughput: float
    name: str
    size_mb: float
    framework: DeepLearningFramework


@dataclass
class BenchmarkOriginalModelResult:
    """The result of the LatencyOriginalModelMeasureOp"""

    latency_seconds: float
    model_outputs: Any


@dataclass
class OptimizeInferenceResult:
    """The result of the OptimizeInferenceOp"""

    original_model: OriginalModel
    hardware_setup: HardwareSetup
    optimized_model: Optional[OptimizedModel]

    @property
    def metric_drop(self) -> Optional[float]:
        if self.optimized_model is None:
            return None
        return self.optimized_model.metric_drop

    @cached_property
    def latency_improvement_rate(self) -> Optional[float]:
        if self.optimized_model is None:
            return None
        if self.optimized_model.latency_seconds == 0:
            return -1
        return (
            self.original_model.latency_seconds
            / self.optimized_model.latency_seconds
        )

    @cached_property
    def throughput_improvement_rate(self) -> Optional[float]:
        if self.optimized_model is None:
            return None
        if self.original_model.throughput == 0:
            return -1
        return self.optimized_model.throughput / self.original_model.throughput

    @cached_property
    def size_improvement_rate(self) -> Optional[float]:
        if self.optimized_model is None:
            return None
        if self.optimized_model.size_mb == 0:
            return 1
        return self.original_model.size_mb / self.optimized_model.size_mb


class InputInfo:
    """Class for storing all the information needed for creating an input
    tensor for AI models.

    Attributes:
        size (tuple): Tuple with the input size (batch size excluded)
        dtype (str): Data type of the tensor.
        min_value (int or float, optional): Min value the tensor elements can
            have.
        max_value (int or float, optional): Max value the tensor elements can
            have.
    """

    def __init__(self, size: Tuple[int, ...], dtype: str, **extra_info):
        self.dtype = DataType(dtype)
        self.size = size
        self.__dict__.update(extra_info)

    def __getattr__(self, item):
        return self.__dict__.get(item)

    def dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith("_")
        }


@dataclass
class DynamicAxisInfo:
    inputs: List[Dict[int, str]]
    outputs: List[Dict[int, str]]

    def dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith("_")
        }

    def retrieve_output_dim(
        self,
        input_shapes: List[Tuple[int, ...]],
        output_idx: int,
        dimension_idx: int,
        default_output_value: int,
    ) -> int:
        output_tag = self.outputs[output_idx][dimension_idx]
        for input_dict, input_shape in zip(self.inputs, input_shapes):
            for key, value in input_dict.items():
                if (
                    isinstance(value, dict) and value.get("name") == output_tag
                ) or value == output_tag:
                    return input_shape[key]
        return default_output_value


@dataclass
class ModelParams:
    batch_size: int
    input_infos: List[InputInfo]
    output_sizes: List[Tuple[int, ...]]
    output_types: List[DataType]
    dynamic_info: Union[DynamicAxisInfo, Dict] = None

    def __post_init__(self):
        if isinstance(self.dynamic_info, dict):
            self.dynamic_info = DynamicAxisInfo(**self.dynamic_info)
        self.input_infos = [
            InputInfo(**x) if isinstance(x, dict) else x
            for x in self.input_infos
        ]
        self.output_types = [DataType(x) for x in self.output_types]

    def dict(self):
        def recursively_dictionarize(element):
            if isinstance(element, list):
                element = [recursively_dictionarize(el) for el in element]
            elif hasattr(element, "dict"):
                element = element.dict()
            return element

        return {
            k: recursively_dictionarize(v)
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }

    @property
    def input_sizes(self):
        for input_info in self.input_infos:
            yield input_info.size


class Device:
    def __init__(self, type: DeviceType, idx: int = 0):
        self.type = type
        self.idx = idx

    @classmethod
    def from_str(cls, string: str) -> "Device":
        if string.startswith("cuda") or string.startswith("gpu"):
            return cls(
                DeviceType.GPU,
                int(string.split(":")[1] if ":" in string else 0),
            )
        elif string.startswith("tpu"):
            return cls(
                DeviceType.TPU,
                int(string.split(":")[1] if ":" in string else 0),
            )

        return cls(DeviceType.CPU)

    def to_torch_format(self) -> str:
        if self.type is DeviceType.GPU:
            return f"cuda:{self.idx}"
        elif self.type is DeviceType.TPU:
            return f"xla:{self.idx}"

        return "cpu"

    def to_tf_format(self) -> str:
        if self.type is DeviceType.GPU:
            return f"GPU:{self.idx}"

        return "CPU"

    def get_total_memory(self) -> int:
        # Return total memory in bytes using nvidia-smi in bytes
        if self.type is not DeviceType.GPU:
            raise Exception("Device type must be GPU")
        else:
            try:
                output = (
                    subprocess.check_output(
                        "nvidia-smi --query-gpu=memory.total "
                        "--format=csv,nounits,noheader",
                        shell=True,
                    )
                    .decode("utf-8")
                    .split()[self.idx]
                )
                return int(output) * 1024 * 1024
            except Exception:
                raise Exception(
                    "Unable to get total memory of device. "
                    "Please make sure nvidia-smi is available."
                )

    def get_free_memory(self) -> int:
        # Return free memory in bytes using nvidia-smi in bytes
        if self.type is not DeviceType.GPU:
            raise Exception("Device type must be GPU")
        else:
            try:
                output = (
                    subprocess.check_output(
                        "nvidia-smi --query-gpu=memory.free "
                        "--format=csv,nounits,noheader",
                        shell=True,
                    )
                    .decode("utf-8")
                    .split()[self.idx]
                )
                return int(output) * 1024 * 1024
            except Exception:
                raise Exception(
                    "Unable to get free memory of device. "
                    "Please make sure nvidia-smi is available."
                )


FRAMEWORK_TO_DATA_TYPE_CONVERSION_DICT = {
    "torch": {
        torch.float16: DataType.FLOAT16,
        torch.float32: DataType.FLOAT32,
        torch.int32: DataType.INT32,
        torch.int64: DataType.INT64,
    },
    "tensorflow": {
        tf.float16: DataType.FLOAT16,
        tf.float32: DataType.FLOAT32,
        tf.int32: DataType.INT32,
        tf.int64: DataType.INT64,
    },
    "numpy": {
        np.float16: DataType.FLOAT16,
        np.float32: DataType.FLOAT32,
        np.int32: DataType.INT32,
        np.int64: DataType.INT64,
    },
}
